/* Copyright (c) 2001 Jim Winstead Jr. See the LICENSE file for details. */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/shape.h>

#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <libgen.h>
#include <time.h>
#include <signal.h>

#define DEFAULT_FONT     "fixed"
#define DEFAULT_FORMAT   "%a %b %d %H:%M:%S %Y"

#define die(x) do { \
          fprintf(stderr,"%s: %s\n",arg0,(x)); \
          exitcode = 1; \
          goto exit; \
        } while (0)

#define white WhitePixel(display,DefaultScreen(display))
#define black BlackPixel(display,DefaultScreen(display))

int exitcode = 0;
char *arg0 = NULL;

Display *display = NULL;
Window window = 0;
int updateinterval = 1;

void update(int sig) {
  XEvent event;
  event.type = Expose; event.xexpose.count = 0;
  XSendEvent(display,window,True,0,(XEvent*)&event);
  XFlush(display);
}

int main(int argc, char **argv) {
  XFontSet fontset = NULL;
  XFontSetExtents *extents = NULL;
  Pixmap mask = 0;
  Colormap colormap = 0;
  GC wgc = NULL, mgc = NULL;
  XEvent event;
  XSetWindowAttributes attributes;
  XRectangle ink, logical;

  char *fontname = NULL, *format = NULL, *color = NULL, *shadow = NULL,
       **missing = NULL, *def_string = NULL, buff[256];
  int c, nmissing, x = 0, y = 0, width = 0, height = 0, offset = 2,
      textcolor = 0, shadowcolor = 0, rightalign = 0;

  arg0 = basename(argv[0]);

  while ((c = getopt(argc,argv,"c:f:g:ho:rs:u:?")) > 0) {
    switch (c) {
      case 'c':
	color = strdup(optarg);
	break;
      case 'f':
	fontname = strdup(optarg);
	break;
      case 'g': {
          int r = XParseGeometry(optarg, &x, &y, &width, &height);
	  if (r&XNegative) die("can't handle a negative x position (use -r?)");
          break;
	}
      case 'o':
	offset = atoi(optarg);
	break;
      case 'r':
	rightalign++;
	break;
      case 's':
	shadow = strdup(optarg);
	break;
      case 'u':
	updateinterval = atoi(optarg);
	break;
      case 'h':
      case '?':
      default:
	fprintf(stderr,"usage: %s [options] [+format]\n", arg0);
	fprintf(stderr,"Display the time.\n");
	fprintf(stderr,"  -c color     Use color for text.\n");
	fprintf(stderr,"  -f font      Use font.\n");
	fprintf(stderr,"  -g geometry  Use geometry.\n");
	fprintf(stderr,"  -o offset    Use offset for shadow.\n");
	fprintf(stderr,"  -r           Right-align the text.\n");
	fprintf(stderr,"  -s color     Use color for shadow.\n");
	fprintf(stderr,"  -u interval  Update every interval seconds.\n");
	fprintf(stderr,"  +format      Use strftime() format.\n");
	exit(0);
    }
  }
  if (optind < argc && argv[optind][0] == '+') {
    format = strdup(argv[optind]+1);
  }

  display = XOpenDisplay(NULL);
  if (!display) die("unable to open display");

  if (!XShapeQueryExtension(display, &c, &c)) /* we don't need the pointers */
    die("shape extension not supported");

  /* load the font set */
  fontset = XCreateFontSet(display, fontname ? fontname : DEFAULT_FONT,
                           &missing, &nmissing, &def_string);
  if (!fontset) die("unable to create font set");

  /* get the extents of the font */
  extents = XExtentsOfFontSet(fontset);

  if (!height)height = extents->max_logical_extent.height + offset;
  if (!y) y = HeightOfScreen(DefaultScreenOfDisplay(display)) - height;
  if (y < 0) y = HeightOfScreen(DefaultScreenOfDisplay(display)) - height + y;

  if (!x) x = 0;
  if (!width) width = WidthOfScreen(DefaultScreenOfDisplay(display)) - x;

  /* create our window */
  attributes.override_redirect = 1;
  window = XCreateWindow(display, DefaultRootWindow(display),
                         x, y, width, height, 0,
			 CopyFromParent, CopyFromParent, CopyFromParent,
			 CWOverrideRedirect,&attributes);
  if (!window) die("unable to create window");
  XStoreName(display,window,arg0);

  /* our drawing surface */
  mask = XCreatePixmap(display,window,width,height,1);
  if (!mask) die("unable to allocate mask for drawing");

  /* our colormap */
  colormap = DefaultColormap(display, DefaultScreen(display));
  if (!colormap) die("unable to get colormap");

  /* grab colors */
  if (color) {
    XColor avail,exact;
    if (XAllocNamedColor(display,colormap,color,&avail,&exact)) {
      textcolor = avail.pixel;
    }
  }
  if (!textcolor) textcolor = white;

  if (shadow) {
    XColor avail,exact;
    if (XAllocNamedColor(display,colormap,color,&avail,&exact)) {
      shadowcolor = avail.pixel;
    }
  }
  if (!shadowcolor) shadowcolor = black;

  /* our graphics context*/
  wgc = XCreateGC(display,window,0,NULL);
  mgc = XCreateGC(display,mask,0,NULL);
  if (!wgc || !mgc) die("unable to create graphics context(s)");

  XSetBackground(display,wgc,black);
  XSetBackground(display,mgc,black);

  /* we want to know about exposures */
  XSelectInput (display, window, ExposureMask|VisibilityChangeMask);

  /* map the window and get things underway */
  XMapRaised(display,window);

  signal(SIGALRM,update);

  for (;;) {
    XNextEvent(display,&event);
    /* don't let the bastards keep us down. */
    if (event.type == VisibilityNotify) {
      XRaiseWindow(display,window);
    }
    /* we only care about the last expose event in a series */
    if (event.type == Expose && event.xexpose.count == 0) {
      /* what time is it? */
      time_t now = time(NULL);
      c = strftime(buff,sizeof(buff),
	           format ? format : DEFAULT_FORMAT,
		   localtime(&now));

      if (rightalign) {
        XmbTextExtents(fontset, buff, c, &ink, &logical);
        x = width - logical.width - offset;
      }

      /* draw the text on the mask (clearing it first) */
      XSetForeground(display,mgc,black);
      XFillRectangle(display,mask,mgc,x,0,width,height);
      XSetForeground(display,mgc,white);
      if (offset)
        XmbDrawString(display,mask,fontset,mgc,x+offset,height,buff,c);
      XmbDrawString(display,mask,fontset,mgc,x,height-offset,buff,c);
      
      /* update our shape mask */
      XShapeCombineMask (display,window,ShapeBounding,0,0,mask,ShapeSet);

      /* draw the text on the window */
      if (offset) {
        XSetForeground(display,wgc,shadowcolor);
        XmbDrawString(display,window,fontset,wgc,x+offset,height,buff,c);
      }
      XSetForeground(display,wgc,textcolor);
      XmbDrawString(display,window,fontset,wgc,x,height-offset,buff,c);

      alarm(updateinterval);
    }
  }

exit:
  /* clean up */
  if(mgc)     XFreeGC(display,mgc);
  if(wgc)     XFreeGC(display,wgc);
  if(colormap)XFreeColormap(display,colormap);
  if(mask)    XFreePixmap(display,mask);
  if(window)  XDestroyWindow(display,window);
  if(missing) XFreeStringList(missing);
  if(fontset) XFreeFontSet(display,fontset);
  if(display) XCloseDisplay(display);
  if(fontname)free(fontname);
  if(shadow)  free(shadow);
  if(color)   free(color);

  exit(exitcode);
}
